//
//  CWXMLParser.m
//  XmlTranslater
//
//  Copyright 2009-2010 Jayway. All rights reserved.
//  Licensed under the Apache License, Version 2.0 (the "License");
//  you may not use this file except in compliance with the License.
//  You may obtain a copy of the License at
//
//  http://www.apache.org/licenses/LICENSE-2.0
//
//  Unless required by applicable law or agreed to in writing, software
//  distributed under the License is distributed on an "AS IS" BASIS,
//  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
//  See the License for the specific language governing permissions and
//  limitations under the License.
//

#import "CWXMLTranslater.h"


@interface CWXMLTranslaterState : NSObject
{
@public
	NSDictionary* translationPlist;
  id currentObject;
  NSString* elementName;
  int nestingCount;
}

-(id)initWithObject:(id)object;

@end

@implementation CWXMLTranslaterState

-(id)initWithObject:(id)object;
{
  self = [super init];
  if (self) {
	  elementName = @"";
		currentObject = [object retain];
  }
  return self;
}

-(void)dealloc;
{
	[currentObject release];
  [super dealloc];
}

@end


@implementation CWXMLTranslater

@synthesize delegate = _delegate;

-(id)initWithTranslationPropertyList:(NSDictionary*)translationPropertyList delegate:(id<CWXMLTranslaterDelegate>)delegate;
{
	self = [self init];
  if (self) {
  	translationPlist = [translationPropertyList copy];
    self.delegate = delegate;
  }
  return self;
}

-(void)dealloc;
{
	[translationPlist release];
	[stateStack release];
  [currentText release];
  [super dealloc];
}


-(NSArray*)translateContentsOfURL:(NSURL*)url;
{
	return [self translateContentsOfURL:url delegate:self.delegate error:NULL];
}

-(NSArray*)translateContentsOfURL:(NSURL*)url delegate:(id<CWXMLTranslaterDelegate>)delegate error:(NSError**)error;
{
  BOOL result = NO;
  didAbort = NO;
  rootObjects = [NSMutableArray array];
  NSLog(@"URL: %@", [url description]);
  NSError* tmpError = nil;
  if (error == NULL) {
    error = &tmpError;
  }
  NSURLRequest* request = [NSURLRequest requestWithURL:url cachePolicy:NSURLRequestUseProtocolCachePolicy timeoutInterval:30];
	NSData* data = [NSURLConnection sendSynchronousRequest:request returningResponse:NULL error:error];
  if (data != nil) {
    xmlParser = [[NSXMLParser alloc] initWithData:data];

    if (xmlParser) {
      [xmlParser setDelegate:self];
      currentDelegate = delegate;
      CWXMLTranslaterState* state = [[CWXMLTranslaterState alloc] init];
      state->translationPlist = translationPlist;
      stateStack = [[NSMutableArray alloc] initWithObjects:state, nil];
      [state release];
      result = [xmlParser parse];
      if (!result) {
        if (didAbort) {
          result = YES;
        } else {
          *error = [xmlParser parserError];          
        }
      }
      [stateStack release];
      stateStack = nil;
      currentDelegate = nil;
      [xmlParser release];
      xmlParser = nil;
    }
  }
  if (result == NO) {
    NSLog(@"Unparsable data: %@", [[[NSString alloc] initWithData:data encoding:NSUTF8StringEncoding] autorelease]);
  	rootObjects = nil;
  }
	return [[rootObjects copy] autorelease];
}

-(void)abortTranslation;
{
  didAbort = YES;
  [xmlParser abortParsing];
}

-(id)willTranslateValueOfClass:(Class)aClass withString:(NSString*)aString forKey:(NSString*)key;
{
  id result = nil;
  if ([currentDelegate respondsToSelector:@selector(xmlTranslater:willTranslateValueOfClass:withString:forKey:)]) {
		result = [currentDelegate xmlTranslater:self willTranslateValueOfClass:aClass withString:aString forKey:key];
  }
  if (result == nil) {
    if (aClass == [NSString class]) {
      return aString;
    } else if (aClass == [NSNumber class]) {
    	result = [NSDecimalNumber decimalNumberWithString:aString];
    } else if (aClass == [NSDate class]) {
      NSDateFormatter* formatter = [[NSDateFormatter alloc] init];
      result = [formatter dateFromString:aString];
      [formatter release];
    } else {
    	result = [[[aClass alloc] initWithString:aString] autorelease];
    }
  }
  return result;
}

-(void)setValue:(id)value forKey:(id)key onObject:(id)target;
{
  if ([key isKindOfClass:[NSArray class]]) {
    value = [self willTranslateValueOfClass:NSClassFromString([key objectAtIndex:1]) withString:value forKey:[key objectAtIndex:0]];
    key = [key objectAtIndex:0];
  } else {
  	value = [self willTranslateValueOfClass:[NSString class] withString:value forKey:key];
  }
  if ([key hasPrefix:@"+"]) {
    key = [key substringFromIndex:1];
    [[target mutableArrayValueForKey:key] addObject:value];
  } else {
	  [target setValue:value forKey:key];
  }
}


-(id)willTranslateObjectOfClass:(Class)aClass forKey:(NSString*)key;
{
  id result = nil;
  if ([currentDelegate respondsToSelector:@selector(xmlTranslater:willTranslateObjectOfClass:forKey:)]) {
		result = [currentDelegate xmlTranslater:self willTranslateObjectOfClass:aClass forKey:key];
  }
  if (result == nil) {
  	result = [[[aClass alloc] init] autorelease];
  }
  return result;
}

- (void)parser:(NSXMLParser *)parser didStartElement:(NSString *)elementName namespaceURI:(NSString *)namespaceURI qualifiedName:(NSString *)qualifiedName attributes:(NSDictionary *)attributeDict;
{
  CWXMLTranslaterState* state = [stateStack lastObject];
  NSDictionary* currentTranslationPlist = ((CWXMLTranslaterState*)[stateStack lastObject])->translationPlist;
	for (NSString* translation in currentTranslationPlist) {
		NSArray* parts = [translation componentsSeparatedByString:@"."];
    if ([parts count] == 1 && [translation isEqualToString:elementName]) {
      id target = [currentTranslationPlist objectForKey:translation];
			if ([target isKindOfClass:[NSDictionary class]]) {
        id currentObject = nil;
        if (![[target objectForKey:@"@dummy"] boolValue]) {
          Class aClass = NSClassFromString([target objectForKey:@"@class"]);
          currentObject = [self willTranslateObjectOfClass:aClass forKey:elementName];
        }
        CWXMLTranslaterState* state = [[CWXMLTranslaterState alloc] initWithObject:currentObject];
        state->elementName = elementName;
        state->translationPlist = target;
        [stateStack addObject:state];
        [state release];
      } else if ([target isEqual:@"@object"]) {
        currentText = [[NSMutableString alloc] init];
        CWXMLTranslaterState* state = [[CWXMLTranslaterState alloc] initWithObject:currentText];
        state->elementName = elementName;
        state->translationPlist = nil;
        [stateStack addObject:state];
        [state release];
      } else {
        if ([state->elementName isEqualToString:elementName]) {
      		state->nestingCount++;
        }
        currentText = [[NSMutableString alloc] init];
      }
    } else {
      NSString* tag = [parts objectAtIndex:0];
      if ([tag length] == 0 || [tag isEqualToString:elementName]) {
        NSString* value = [attributeDict objectForKey:[parts objectAtIndex:1]];
        id key = [currentTranslationPlist objectForKey:translation];
        [self setValue:value forKey:key onObject:state->currentObject];
      }
    }
  }
}

- (void)parser:(NSXMLParser *)parser foundCharacters:(NSString *)string;
{
	[currentText appendString:string];
}

- (void)parser:(NSXMLParser *)parser didEndElement:(NSString *)elementName namespaceURI:(NSString *)namespaceURI qualifiedName:(NSString *)qName;
{
	CWXMLTranslaterState* state = [stateStack lastObject];
  if ([state->elementName isEqualToString:elementName]) {
  	if (state->nestingCount > 0) {
    	state->nestingCount--;
    } else {
      NSString* key = nil;
      if ([state->translationPlist isKindOfClass:[NSDictionary class]]) {
        key = [state->translationPlist objectForKey:@"@key"];
      }
      if (key) {
				CWXMLTranslaterState* prevState = nil;
        for (int i = 2; YES; i++) {
        	prevState = [stateStack objectAtIndex:[stateStack count] - i];
          if (prevState->currentObject) {
          	break;
          }
        }
        [self setValue:state->currentObject forKey:key onObject:prevState->currentObject];
      } else if (state->currentObject) {
        id object = state->currentObject;
				if ([currentDelegate respondsToSelector:@selector(xmlTranslater:didTranslateObject:forKey:)]) {
        	object = [currentDelegate xmlTranslater:self didTranslateObject:object forKey:elementName];
        }
        if (object != nil) {
	        [rootObjects addObject:state->currentObject];
        }
      }
      [stateStack removeLastObject];
    }
  } else {
    for (NSString* tag in state->translationPlist) {
      if ([tag isEqualToString:elementName]) {
        id key = [state->translationPlist objectForKey:tag];
        [self setValue:currentText forKey:key onObject:state->currentObject];
      }
    }
  }
  [currentText release];
  currentText = nil;
}

@end
